from __future__ import annotations

from functools import cached_property, partial
from itertools import permutations
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple

from . import utilities
from .board_state import BoardState
from .config import Config
from .sokoban_plus import SokobanPlus

if TYPE_CHECKING:
    from .board_graph import BoardGraph


class CellAlreadyOccupiedError(RuntimeError):
    pass


class BoxGoalSwitchError(RuntimeError):
    pass


class BoardManager:
    """Memoizes, tracks and updates positions of all pieces.

    - Provides efficient means to inspect positions of pushers, boxes and goals. To
      understand how this works, we need to have a way of identifying individual
      pushers, boxes and goals. :class:`.BoardManager` does that by assigning
      numerical ID to each individual piece. This ID can then be used to refer to
      that piece in various contexts.

      How are piece IDs assigned? We start scanning game board from top left corner to
      the right, row by row. First encountered box will get ``box.id =
      Config.DEFAULT_PIECE_ID``, second one ``box.id = Config.DEFAULT_PIECE_ID + 1``,
      etc... Same goes for pushers and goals.

      .. image:: /images/assigning_ids.png
          :alt: Assigning board elements' IDs

    - Provides efficient means of pieces movement. Ie. we can move pushers and boxes
      and :class:`.BoardManager` will update internal state and board cells.

      This movement preserves piece IDs in context of board state changes. To illustrate,
      let's assume we create :class:`.BoardManager` from board with two pushers one
      above the other. After then we edit the board, placing pusher ID 2 in row above
      pusher ID 1. Finally, we create another instance of :class:`.BoardManager`. If we
      now inspect pusher IDs in first and second :class:`.BoardManager` instance, they
      will be different. Have we used movement methods instead of board editing, these
      IDs would be preserved:

      .. |img1| image:: /images/movement_vs_transfer1.png
      .. |img2| image:: /images/movement_vs_transfer2.png
      .. |img3| image:: /images/movement_vs_transfer3.png

      +------------------+------------------+------------------+
      | 1) Initial board | 2) Edited board  | 3) Box ID:1 moved|
      +------------------+------------------+------------------+
      |      |img1|      |      |img2|      |      |img3|      |
      +------------------+------------------+------------------+

    Note:
        Movement methods here are just for state and board cell updates, they don't
        implement full game logic. For game logic see :class:`.Mover`
    """

    def __init__(
        self,
        board: BoardGraph,
        boxorder: Optional[str] = None,
        goalorder: Optional[str] = None,
    ):
        self._board = board
        self._boxes = utilities.Flipdict()
        self._goals = utilities.Flipdict()
        self._pushers = utilities.Flipdict()
        self._walls: List[int] = []

        pusher_id = box_id = goal_id = Config.DEFAULT_PIECE_ID

        for position in range(0, board.vertices_count):
            cell = board[position]

            if cell.has_pusher:
                self._pushers[pusher_id] = position
                pusher_id += 1

            if cell.has_box:
                self._boxes[box_id] = position
                box_id += 1

            if cell.has_goal:
                self._goals[goal_id] = position
                goal_id += 1

            if cell.is_wall:
                self._walls.append(position)

        self._sokoban_plus = SokobanPlus(
            pieces_count=len(self._boxes), boxorder=boxorder, goalorder=goalorder
        )

    def __str__(self):
        prefix = (len(self.__class__.__name__) + 2) * " "
        return "\n".join(
            [
                "<{} pushers: {},".format(
                    self.__class__.__name__, self.pushers_positions
                ),
                prefix + "boxes: {},".format(self.boxes_positions),
                prefix + "goals: {},".format(self.goals_positions),
                prefix
                + "walls: {},".format(
                    self.walls_positions
                    if len(self.walls_positions) <= 10
                    else "[{}, ...]".format(
                        ", ".join(str(w) for w in self.walls_positions[:10])
                    )
                ),
                prefix + "boxorder: '{}',".format(self.boxorder or ""),
                prefix + "goalorder: '{}',".format(self.goalorder or "") + ">",
            ]
        )

    @property
    def board(self) -> BoardGraph:
        return self._board

    @property
    def walls_positions(self) -> List[int]:
        return self._walls

    # --------------------------------------------------------------------------
    # Pushers
    # --------------------------------------------------------------------------

    @cached_property
    def pushers_count(self) -> int:
        return len(self._pushers)

    @cached_property
    def pushers_ids(self) -> List[int]:
        """
        IDs of all pushers on board.
        """
        return list(self._pushers.keys())

    @property
    def pushers_positions(self) -> Dict[int, int]:
        """
        Mapping of pushers' IDs to the corresponding board positions, ie. ::

            {1: 42, 2: 24}
        """
        return dict(self._pushers)

    def pusher_position(self, pusher_id: int) -> int:
        """
        Raises:
            :exc:`KeyError`: No pusher with ID ``pusher_id``
        """
        try:
            return self._pushers[pusher_id]
        except KeyError:
            raise KeyError("No pusher with ID: {0}".format(pusher_id))

    def pusher_id_on(self, position: int) -> int:
        """
        Raises:
            :exc:`KeyError`: No pusher on ``position``
        """
        try:
            return self._pushers.flip[position]
        except KeyError:
            raise KeyError("No pusher on position: {0}".format(position))

    def has_pusher(self, pusher_id: int) -> bool:
        return pusher_id in self._pushers

    def has_pusher_on(self, position: int) -> bool:
        return position in self._pushers.flip

    def _pusher_moved(self, old_position: int, to_new_position: int):
        """Subclass hook used to notify it that we modified pusher position."""
        pass

    def move_pusher_from(self, old_position: int, to_new_position: int):
        """
        Updates board state and board cells with changed pusher position.

        Raises:
            :exc:`KeyError`: there is no pusher on ``old_position``
            :exc:`CellAlreadyOccupiedError`: there is an obstacle (
                wall/box/another pusher) on ``to_new_position``
        """
        if old_position == to_new_position:
            return

        dest_cell = self._board[to_new_position]
        if not dest_cell.can_put_pusher_or_box:
            raise CellAlreadyOccupiedError(
                "Pusher ID: {0} ".format(self.pusher_id_on(old_position))
                + "can't be placed in position {0} occupied by '{1}'".format(
                    to_new_position, dest_cell
                )
            )

        self._pushers[self._pushers.flip[old_position]] = to_new_position
        self._board[old_position].remove_pusher()
        dest_cell.put_pusher()

        self._pusher_moved(old_position, to_new_position)

    def move_pusher(self, pusher_id: int, to_new_position: int):
        """
        Updates board state and board cells with changed pusher position.

        Raises:
            :exc:`KeyError`: there is no pusher with ID ``pusher_id``
            :exc:`CellAlreadyOccupiedError`: there is a pusher already on
                ``to_new_position``

        Note:
            Allows placing a pusher onto position occupied by box. This is for cases
            when we switch box/goals positions in reverse solving mode. In this
            situation it is legal for pusher to end up standing on top of the box.
            Game rules say that for these situations, first move(s) must be jumps.

        Warning:
            It doesn't verify if ``to_new_position`` is valid on-board position.
        """
        self.move_pusher_from(self._pushers[pusher_id], to_new_position)

    # --------------------------------------------------------------------------
    # Boxes
    # --------------------------------------------------------------------------

    @cached_property
    def boxes_count(self) -> int:
        return len(self._boxes)

    @cached_property
    def boxes_ids(self) -> List[int]:
        """
        IDs of all boxes on board.
        """
        return list(self._boxes.keys())

    @property
    def boxes_positions(self) -> Dict[int, int]:
        """
        Mapping of boxes' IDs to the corresponding board positions, ie. ::

            {1: 42, 2: 24}
        """
        return dict(self._boxes)

    def box_position(self, box_id: int) -> int:
        """
        Raises:
            :exc:`KeyError`: No box with ID ``box_id``
        """
        try:
            return self._boxes[box_id]
        except KeyError:
            raise KeyError("No box with ID: {0}".format(box_id))

    def box_id_on(self, position: int) -> int:
        """
        Raises:
            :exc:`KeyError`: No box on ``position``
        """
        try:
            return self._boxes.flip[position]
        except KeyError:
            raise KeyError("No box on position: {0}".format(position))

    def has_box(self, box_id: int) -> bool:
        return box_id in self._boxes

    def has_box_on(self, position: int) -> bool:
        return position in self._boxes.flip

    def _box_moved(self, old_position: int, to_new_position: int):
        """Subclass hook used to notify it that we modified box position."""
        pass

    def move_box_from(self, old_position: int, to_new_position: int):
        """
        Updates board state and board cells with changed box position.

        Raises:
            :exc:`KeyError`: there is no box on ``old_position``
            :exc:`CellAlreadyOccupiedError`: there is an obstacle ( wall/box/pusher)
                on ``to_new_position``
        """
        if old_position == to_new_position:
            return

        dest_cell = self._board[to_new_position]
        if not dest_cell.can_put_pusher_or_box:
            raise CellAlreadyOccupiedError(
                "Box ID: {0} ".format(self.box_id_on(old_position))
                + "can't be placed in position {0} occupied by '{1}'".format(
                    to_new_position, dest_cell
                )
            )

        self._boxes[self._boxes.flip[old_position]] = to_new_position
        self._board[old_position].remove_box()
        dest_cell.put_box()

        self._box_moved(old_position, to_new_position)

    def move_box(self, box_id: int, to_new_position: int):
        """
        Updates board state and board cells with changed box position.

        Raises:
            :exc:`KeyError`: there is no box on ``old_position``
            :exc:`CellAlreadyOccupiedError`: there is an obstacle ( wall/box/another
                pusher) on ``to_new_position``
        """
        self.move_box_from(self._boxes[box_id], to_new_position)

    # --------------------------------------------------------------------------
    # Goals
    # --------------------------------------------------------------------------

    @cached_property
    def goals_count(self) -> int:
        return len(self._goals)

    @cached_property
    def goals_ids(self) -> List[int]:
        """
        IDs of all goals on board.
        """
        return list(self._goals.keys())

    @property
    def goals_positions(self) -> Dict[int, int]:
        """
        Mapping of boxes' IDs to the corresponding board positions, ie. ::

            {1: 42, 2: 24}
        """
        return dict(self._goals)

    def goal_position(self, goal_id: int) -> int:
        """
        Raises:
            :exc:`KeyError`: No goal with ID ``goal_id``
        """
        try:
            return self._goals[goal_id]
        except KeyError:
            raise KeyError("No goal with ID: {0}".format(goal_id))

    def goal_id_on(self, position: int) -> int:
        """
        ID of goal on position.

        Raises:
            :exc:`KeyError`: No goal on ``position``
        """
        try:
            return self._goals.flip[position]
        except KeyError:
            raise KeyError("No goal on position: {0}".format(position))

    def has_goal(self, goal_id: int) -> bool:
        return goal_id in self._goals

    def has_goal_on(self, position: int) -> bool:
        return position in self._goals.flip

    # --------------------------------------------------------------------------
    # Sokoban+
    # --------------------------------------------------------------------------

    def box_plus_id(self, box_id: int) -> int:
        """
        See Also:
            :meth:`.SokobanPlus.box_plus_id`
        """
        return self._sokoban_plus.box_plus_id(box_id)

    def goal_plus_id(self, goal_id: int) -> int:
        """
        See Also:
            :meth:`.SokobanPlus.goal_plus_id`
        """
        return self._sokoban_plus.goal_plus_id(goal_id)

    @property
    def boxorder(self) -> str:
        """
        See Also:
            `.SokobanPlus.boxorder`
        """
        return self._sokoban_plus.boxorder

    @boxorder.setter
    def boxorder(self, rv):
        self._sokoban_plus.boxorder = rv

    @property
    def goalorder(self) -> str:
        """
        See Also:
            `.SokobanPlus.goalorder`
        """
        return self._sokoban_plus.goalorder

    @goalorder.setter
    def goalorder(self, rv):
        self._sokoban_plus.goalorder = rv

    @property
    def is_sokoban_plus_enabled(self) -> bool:
        return self._sokoban_plus.is_enabled

    def enable_sokoban_plus(self):
        """
        Enables using Sokoban+ rules for this board.

        Enabling these, changes victory condition for given board.

        See Also:
            :class:`.SokobanPlus`
        """
        self._sokoban_plus.is_enabled = True

    def disable_sokoban_plus(self):
        """
        Disables using Sokoban+ rules for this board.

        See Also:
            :class:`.SokobanPlus`
        """
        self._sokoban_plus.is_enabled = False

    @property
    def is_sokoban_plus_valid(self) -> bool:
        """
        Validates current set of Sokoban+ rules.

        See Also:
            :class:`.SokobanPlus`
        """
        return self._sokoban_plus.is_valid

    # --------------------------------------------------------------------------
    # Other
    # --------------------------------------------------------------------------

    @property
    def is_solved(self) -> bool:
        if self.boxes_count != self.goals_count:
            return False

        retv = True
        for box_id, box_position in self._boxes.items():
            retv = (
                retv
                and self.has_goal_on(box_position)
                and (
                    self.box_plus_id(box_id)
                    == self.goal_plus_id(self.goal_id_on(box_position))
                )
            )

            if not retv:
                break

        return retv

    def solutions(self) -> Iterable[BoardState]:
        """
        Generator for all configurations of boxes that result in solved board.

        Note:
            Resultset depends on `.BoardManager.is_sokoban_plus_enabled`.
        """
        if self.boxes_count != self.goals_count:
            return []

        def is_valid_solution(boxes_positions):
            retv = True
            for index, box_position in enumerate(boxes_positions):
                box_id = index + Config.DEFAULT_PIECE_ID
                box_plus_id = self.box_plus_id(box_id)
                goal_id = self.goal_id_on(box_position)
                goal_plus_id = self.goal_plus_id(goal_id)

                retv = retv and (box_plus_id == goal_plus_id)
                if not retv:
                    break
            return retv

        for boxes_positions in permutations(self._goals.values()):
            if is_valid_solution(boxes_positions):
                yield BoardState(
                    boxes_positions=list(boxes_positions), pushers_positions=[]
                )

    def _box_goal_pairs(self) -> Iterable[Tuple[int, int]]:
        """
        Finds a list of paired (box_id, goal_id) tuples.

        If Sokoban+ is enabled, boxes and goals are paired by Sokoban+ IDs, otherwise
        they are paired by regular IDs
        """
        if self.boxes_count != self.goals_count:
            return []

        def is_box_goal_pair(box, goal_id):
            if self.is_sokoban_plus_enabled:
                return self.box_plus_id(box[1]) == self.goal_plus_id(goal_id)
            return box[1] == goal_id

        boxes_todo = list(self.boxes_ids)
        goals_ids = list(self.goals_ids)
        for goal_id in goals_ids:
            predicate = partial(is_box_goal_pair, goal_id=goal_id)
            index, box_id = next(filter(predicate, enumerate(boxes_todo)))
            yield box_id, goal_id
            del boxes_todo[index]

    def switch_boxes_and_goals(self):
        """Switches positions of boxes and goals pairs."""
        if self.boxes_count != self.goals_count:
            raise BoxGoalSwitchError(
                "Unable to switch boxes and goals - counts are not the same"
            )

        for box_id, goal_id in self._box_goal_pairs():
            old_box_position = self._boxes[box_id]
            old_goal_position = self._goals[goal_id]

            if old_box_position != old_goal_position:
                # If there is a pusher on goal, we have to remove it before
                # we put a box there
                moved_pusher_id = None
                if self.has_pusher_on(old_goal_position):
                    moved_pusher_id = self.pusher_id_on(old_goal_position)
                    self._pushers[moved_pusher_id] = -1
                    self._board[old_goal_position].remove_pusher()

                self._goals[goal_id] = old_box_position
                self._board[old_goal_position].remove_goal()
                self._board[old_box_position].put_goal()

                self._boxes[box_id] = old_goal_position
                self._board[old_box_position].remove_box()
                self._board[old_goal_position].put_box()
                self._box_moved(old_box_position, old_goal_position)

                if moved_pusher_id is not None:
                    # There was pusher on former goal cell and was deleted
                    # We now put it back on board, onto new goal position
                    # Net result is that box and goal switched places and
                    # pusher moved to new goal position.
                    self._pushers[moved_pusher_id] = old_box_position
                    self._board[old_box_position].put_pusher()
                    self._pusher_moved(old_goal_position, old_box_position)

    @property
    def is_playable(self) -> bool:
        return (
            self.pushers_count > 0
            and self.boxes_count == self.goals_count
            and self.boxes_count > 0
            and self.goals_count > 0
        )

    @property
    def state(self) -> BoardState:
        pushers_positions = self.pushers_positions
        boxes_positions = self.boxes_positions
        return BoardState(
            pushers_positions=[
                pushers_positions[pusher_id]
                for pusher_id in sorted(pushers_positions.keys())
            ],
            boxes_positions=[
                boxes_positions[box_id] for box_id in sorted(boxes_positions.keys())
            ],
        )



import pytest

from sokoenginepy.game import SokobanPlus, SokobanPlusDataError


@pytest.fixture
def sokoban_plus():
    return SokobanPlus(pieces_count=5, boxorder="42 24 4 2", goalorder="2 24 42 4")


class DescribeSokobanPlus:
    class Describe_init:
        def it_doesnt_validate_on_init(self):
            raised = False
            try:
                sokoban_plus = SokobanPlus(1, "foo bar", "4 2")

            except Exception:
                raised = True

            assert not raised
            assert not sokoban_plus.is_valid

        def it_always_creates_disabled_instance(self):
            sokoban_plus = SokobanPlus(42, "foo bar", "4 2")
            assert not sokoban_plus.is_enabled

    class Describe_boxorder_and_goalorder:
        def test_they_return_order_object_from_init_if_sokoban_plus_is_not_valid(
            self, sokoban_plus
        ):
            sokoban_plus = SokobanPlus(5, "1 2 3 4 5 6", "6 5 4 3 2 1")
            assert sokoban_plus.boxorder == "1 2 3 4 5 6"
            assert sokoban_plus.goalorder == "6 5 4 3 2 1"
            sokoban_plus = SokobanPlus(5, "foo", "bar")
            assert sokoban_plus.boxorder == "foo"
            assert sokoban_plus.goalorder == "bar"

        def test_they_return_parsed_order_string_if_sokoban_plus_is_valid_and_enabled(
            self,
        ):
            sokoban_plus = SokobanPlus(5, "1 2 0 3 0", "3 1 0 2 0 0 0")
            assert sokoban_plus.boxorder == "1 2 0 3 0"
            assert sokoban_plus.goalorder == "3 1 0 2 0 0 0"
            sokoban_plus.is_enabled = True
            assert sokoban_plus.boxorder == "1 2 0 3"
            assert sokoban_plus.goalorder == "3 1 0 2"

        def test_setters_disable_and_invalidate_instance(self):
            sokoban_plus = SokobanPlus(5, "1 2 0 3 0 ", "3 1 0 2 0 0 0")
            sokoban_plus.is_enabled = True
            assert sokoban_plus.is_valid is True
            assert sokoban_plus.is_enabled is True
            sokoban_plus.boxorder = "foo"
            assert sokoban_plus.boxorder == "foo"
            assert sokoban_plus.is_enabled is False
            assert sokoban_plus.is_valid is False

            sokoban_plus = SokobanPlus(5, "1 2 0 3 0 ", "3 1 0 2 0 0 0")
            sokoban_plus.is_enabled = True
            assert sokoban_plus.is_valid is True
            assert sokoban_plus.is_enabled is True
            sokoban_plus.goalorder = "foo"
            assert sokoban_plus.goalorder == "foo"
            assert sokoban_plus.is_enabled is False
            assert sokoban_plus.is_valid is False

        def test_setters_dont_disable_and_invalidate_if_setting_to_equal_value(self):
            sokoban_plus = SokobanPlus(5, "1 2 0 3 0", "3 1 0 2 0 0 0")
            sokoban_plus.is_enabled = True
            assert sokoban_plus.is_valid is True
            assert sokoban_plus.is_enabled is True

            sokoban_plus.boxorder = "1 2 0 3 0"
            assert sokoban_plus.boxorder == "1 2 0 3"
            assert sokoban_plus.is_valid is True
            assert sokoban_plus.is_enabled is True

            sokoban_plus.goalorder = "3 1 0 2 0 0 0"
            assert sokoban_plus.goalorder == "3 1 0 2"
            assert sokoban_plus.is_valid is True
            assert sokoban_plus.is_enabled is True

    class Describe_getter_for_box_plus_id:
        def it_returns_default_plus_id_for_disabled_sokoban_plus(self, sokoban_plus):
            assert sokoban_plus.box_plus_id(42) == SokobanPlus.DEFAULT_PLUS_ID

        def it_returns_box_plus_id_for_enabled_sokoban_plus(self, sokoban_plus):
            sokoban_plus.is_enabled = True
            assert sokoban_plus.box_plus_id(2) == int(sokoban_plus.boxorder.split()[1])

    class Describe_getter_for_goal_plus_id:
        def it_returns_default_plus_id_for_disabled_sokoban_plus(self, sokoban_plus):
            assert sokoban_plus.goal_plus_id(42) == SokobanPlus.DEFAULT_PLUS_ID

        def it_returns_goal_plus_id_for_enabled_sokoban_plus(self, sokoban_plus):
            sokoban_plus.is_enabled = True
            assert sokoban_plus.goal_plus_id(2) == int(
                sokoban_plus.goalorder.split()[1]
            )

    class When_it_is_set_to_enabled:
        def it_parses_init_data(self, sokoban_plus):
            assert not sokoban_plus.is_validated
            sokoban_plus.is_enabled = True
            assert sokoban_plus.is_validated

        def it_raises_on_invalid_data(self):
            sokoban_plus = SokobanPlus(5, "ZOMG!", "YOLO")
            with pytest.raises(SokobanPlusDataError):
                sokoban_plus.is_enabled = True

        def it_enables_sokoban_plus(self, sokoban_plus):
            assert not sokoban_plus.is_enabled
            sokoban_plus.is_enabled = True
            assert sokoban_plus.is_enabled

    class Describe_legacy_plus_id_handling:
        def it_converts_legacy_plus_ids_silently_if_piece_count_is_smaller_than_legacy_plus_id(
            self,
        ):
            sokoban_plus = SokobanPlus(
                pieces_count=10,
                boxorder=f"1 5 42 {SokobanPlus.LEGACY_DEFAULT_PLUS_ID} 23",
                goalorder=f"1 5 42 {SokobanPlus.LEGACY_DEFAULT_PLUS_ID} 23",
            )
            sokoban_plus.is_enabled = True
            assert sokoban_plus.boxorder == "1 5 42 {0} 23".format(
                SokobanPlus.DEFAULT_PLUS_ID
            )
            assert sokoban_plus.goalorder == "1 5 42 {0} 23".format(
                SokobanPlus.DEFAULT_PLUS_ID
            )

        def it_doesnt_convert_legacy_plus_ids_if_piece_count_is_greater_than_legacy_plus_id(
            self, sokoban_plus
        ):
            sokoban_plus = SokobanPlus(
                pieces_count=SokobanPlus.LEGACY_DEFAULT_PLUS_ID,
                boxorder=f"1 5 42 {SokobanPlus.LEGACY_DEFAULT_PLUS_ID} 23",
                goalorder=f"1 5 42 {SokobanPlus.LEGACY_DEFAULT_PLUS_ID} 23",
            )
            sokoban_plus.is_enabled = True
            assert sokoban_plus.boxorder == "1 5 42 {0} 23".format(
                SokobanPlus.LEGACY_DEFAULT_PLUS_ID
            )
            assert sokoban_plus.goalorder == "1 5 42 {0} 23".format(
                SokobanPlus.LEGACY_DEFAULT_PLUS_ID
            )

    class Describe_is_valid:
        def it_validates_all_plus_ids(self):
            sokoban_plus = SokobanPlus(3, "1 2 3", "3 2 1")
            assert sokoban_plus.is_valid
            assert len(sokoban_plus.errors) == 0

            sokoban_plus = SokobanPlus(3, "-1 -2 -3", "3 2 1")
            assert not sokoban_plus.is_valid
            assert len(sokoban_plus.errors) > 0

        def it_validates_piece_count_is_greater_or_equal_than_zero(
            self, is_using_native
        ):
            sokoban_plus = SokobanPlus(42, "1 2 3", "3 2 1")
            assert sokoban_plus.is_valid

            if is_using_native:
                with pytest.raises(TypeError):
                    sokoban_plus = SokobanPlus(-42, "1 2 3", "3 2 1")
            else:
                sokoban_plus = SokobanPlus(-42, "1 2 3", "3 2 1")
                assert not sokoban_plus.is_valid

        def it_validates_ids_lengths_are_equal_to_piece_count(self):
            sokoban_plus = SokobanPlus(4, "1 2 3", "3 2 1")
            assert sokoban_plus.is_valid

            sokoban_plus = SokobanPlus(2, "1 2 3", "3 2 1")
            assert not sokoban_plus.is_valid

        def it_validates_there_are_not_to_many_ids(self):
            sokoban_plus = SokobanPlus(
                goalorder="1 2 3 4 5 6", boxorder="6 5 4 3 2 1", pieces_count=5
            )
            assert not sokoban_plus.is_valid

        def it_validates_same_id_set_is_defined_for_both_piece_types(self):
            sokoban_plus = SokobanPlus(
                goalorder="1 2 3 4 5 6", boxorder="5 4 3 2 1", pieces_count=6
            )
            assert not sokoban_plus.is_valid

        def it_allows_that_there_are_duplicate_plus_ids_for_same_piece_type(self):
            sokoban_plus = SokobanPlus(
                goalorder="1 1 2 2 3 3", boxorder="2 2 3 3 1 1", pieces_count=10
            )
            assert sokoban_plus.is_valid



# http://code.activestate.com/recipes/576968/
# Created by Francis Carr
# Licensed under the MIT License

_NOTFOUND = object()


class Flipdict(dict):
    """An injective (one-to-one) python dict.  Ensures that each key maps
    to a unique value, and each value maps back to that same key.

    A Flipdict is a dict, implementing the dict API::
        >>> fd = Flipdict({1: 'one', 2: 'two', 3: 'three'})
        >>> fd[1]
        'one'
        >>> fd.setdefault(4, 'four')
        'four'
        >>> fd
        Flipdict({1: 'one', 2: 'two', 3: 'three', 4: 'four'})

    Each Flipdict has a "flip" attribute, its inverse mapping::
        >>> fd is fd.flip.flip
        True
        >>> sorted(fd.flip.items())
        [('four', 4), ('one', 1), ('three', 3), ('two', 2)]

    Just like a dict, remapping a key will erase the old mapping.
    IMPORTANT: Suppose a Flipdict maps key K to value V. Then mapping a
    *new* key to the value V will raise KeyError::
        >>> fd[4444] = 'four'
        Traceback (most recent call last):
        ....
        KeyError: "(key,val) would erase mapping for value 'four'"

    This avoids a subtle situation.  The alternative is to silently erase
    the existing pair (K,V)!  Instead, explicitly use the "flip" attribute
    like a python dict::
        >>> fd.flip['four'] = 4444
        >>> fd
        Flipdict({1: 'one', 2: 'two', 3: 'three', 4444: 'four'})
    """

    def __init__(self, *args, **kw):
        """Cf. dict.__init__

        Examples::
            >>> fd = Flipdict( zip((1, 2, 3), ('one', 'two', 'three')) )
            >>> fd == Flipdict({1: 'one', 2: 'two', 3: 'three'})
            True
            >>> fd == Flipdict(one=1, two=2, three=3).flip
            True

        IMPORTANT: Mapping many keys to the same value will raise KeyError.
        """
        self._flip = dict.__new__(self.__class__)
        setattr(self._flip, "_flip", self)
        for key, val in dict(*args, **kw).items():
            self[key] = val

    @property
    def flip(self):
        """The inverse mapping.

        Examples::
            >>> fd = Flipdict({1: 'one', 2: 'two', 3: 'three'})
            >>> fd.flip['two']
            2
            >>> fd is fd.flip.flip
            True
        """
        return self._flip

    # { Non-mutating methods that are NOT delegated to the dict superclass.

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, dict(self))

    __str__ = __repr__

    def copy(self):
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, keys, value=None):
        """
        NOTE: This method is mostly useless for a Flipdict, because
        each key must map to a UNIQUE value.
        """
        return cls(dict.fromkeys(keys, value))

    # }

    # { Mutating methods.  These must keep the inverse mapping in sync!

    def __setitem__(self, key, val):
        """self[key] = val (and self.flip[val] = key)

        Remapping a key will erase the old mapping (just like a dict)::
            >>> fd = Flipdict(one=1, two=2)
            >>> fd['two'] = 22
            >>> fd
            Flipdict({'one': 1, 'two': 22})
            >>> fd.flip
            Flipdict({1: 'one', 22: 'two'})

        Just like a dict, remapping a key will erase the old mapping.
        IMPORTANT: Suppose a Flipdict maps key K to value V. Then mapping a
        *new* key to the value V will raise KeyError::
            >>> fd['twenty-two'] = 22
            Traceback (most recent call last):
            ....
            KeyError: '(key,val) would erase mapping for value 22'

        This avoids a subtle situation.  The alternative is to silently
        erase the existing pair (K,V)!  Instead, explicitly use the "flip"
        attribute like a python dict::
            >>> fd.flip[22] = 'twenty-two'
            >>> sorted(fd.items())
            [('one', 1), ('twenty-two', 22)]
        """
        k = self._flip.get(val, _NOTFOUND)
        if not (k is _NOTFOUND or k == key):
            raise KeyError("(key,val) would erase mapping for value %r" % val)

        v = self.get(key, _NOTFOUND)
        if v is not _NOTFOUND:
            dict.__delitem__(self._flip, v)

        dict.__setitem__(self, key, val)
        dict.__setitem__(self._flip, val, key)

    def setdefault(self, key, default=None):
        # Copied from python's UserDict.DictMixin code.
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def update(self, other=None, **kwargs):
        # Copied from python's UserDict.DictMixin code.
        # Make progressively weaker assumptions about "other"
        if other is None:
            pass
        elif hasattr(other, "iteritems"):  # iteritems saves memory and lookups
            for k, v in other.items():
                self[k] = v
        elif hasattr(other, "keys"):
            for k in list(other.keys()):
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)

    def __delitem__(self, key):
        val = dict.pop(self, key)
        dict.__delitem__(self._flip, val)

    def pop(self, key, *args):
        val = dict.pop(self, key, *args)
        dict.__delitem__(self._flip, val)
        return val

    def popitem(self):
        key, val = dict.popitem(self)
        dict.__delitem__(self._flip, val)
        return key, val

    def clear(self):
        dict.clear(self)
        dict.clear(self._flip)


def makepair(*args, **kw):
    """Returns a pair:  the Flipdict(*args, **kw) and its inverse.

    Example::
        >>> chr2num, num2chr = makepair({'a':1, 'b':2, 'c':3})
        >>> chr2num['z'] = 26
        >>> list(num2chr.items())
        [(1, 'a'), (2, 'b'), (3, 'c'), (26, 'z')]
    """
    fd = Flipdict(*args, **kw)
    return fd, fd.flip


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import shlex
import sys
from datetime import datetime
from pathlib import Path

_SELF_DIR = Path(__file__).parent.absolute().resolve()

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "src")))

nitpicky = True


# -- ReadTheDocs stuff -------------------------------------------------------

on_rtd = os.environ.get("READTHEDOCS", None) == "True"


# -- Project information -----------------------------------------------------

project = "sokoenginepy"
copyright = (
    ", ".join(str(y) for y in range(2017, datetime.now().year + 1))
    + ", Tomislav Adamic"
)
author = "Tomislav Adamic"
version = release = "0.6.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.ifconfig",
    # "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "m2r2",
    # "breathe",
]

needs_sphinx = "4.4"

# autodoc_type_aliases = {
#     "AnyTessellation": "sokoenginepy.AnyTessellation",
#     "TessellationOrDescription": "sokoenginepy.TessellationOrDescription",
# }
autodoc_typehints_format = "short"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
source_suffix = {".rst": "restructuredtext", ".md": "markdown"}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "src/libsokoengine",
    "src/sokoenginepyext",
    "src/utilities",
]

default_role = "py:obj"
add_module_names = False
pygments_style = "sphinx"
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# autodoc_member_order = 'bysource'
autodoc_member_order = "groupwise"

# -- Options for Doxygen C++ input -------------------------------------------

# breathe_projects = {"libsokoengine": str(_SELF_DIR / "_doxygen" / "xml")}
# breathe_default_project = "libsokoengine"



from __future__ import annotations

import enum
from typing import Final, List


class Direction(enum.Enum):
    """Directions of movement."""

    UP = 0
    NORTH_EAST = 1
    RIGHT = 2
    SOUTH_EAST = 3
    DOWN = 4
    SOUTH_WEST = 5
    LEFT = 6
    NORTH_WEST = 7

    def __repr__(self):
        return "Direction." + self.name

    @property
    def opposite(self) -> Direction:
        return _OPPOSITE[self.value]


_OPPOSITE: List[Direction] = [
    Direction.DOWN,
    Direction.SOUTH_WEST,
    Direction.LEFT,
    Direction.NORTH_WEST,
    Direction.UP,
    Direction.NORTH_EAST,
    Direction.RIGHT,
    Direction.SOUTH_EAST,
]


class GraphType(enum.Enum):
    """
    Type of board graph.
    """

    DIRECTED = 0
    DIRECTED_MULTI = 1


class Config:
    MAX_WIDTH: Final[int] = 4096
    MAX_HEIGHT: Final[int] = 4096
    MAX_POS: Final[int] = MAX_WIDTH * MAX_HEIGHT

    NULL_ID: Final[int] = 0

    #: All pieces on board (pushers, boxes) and goals are assigned ID to be used for
    #: referring them.
    #:
    #: Piece ids are assigned by numbering each type of piece or goal starting with
    #: ``DEFAULT_PIECE_ID``
    #:
    #: .. image:: /images/assigning_ids.png
    #:     :alt: Assigning IDs to pieces and goals
    #:
    #: Once assigned, id of a piece doesn't change by either moving piece or
    #: applying board view transformations.
    DEFAULT_PIECE_ID: Final[int] = 1



from __future__ import annotations

from typing import Optional

from .puzzle import Puzzle
from .snapshot import Snapshot


class SokobanSnapshot(Snapshot):
    def __init__(self, moves_data: str = ""):
        from ..game import Tessellation

        super().__init__(tessellation=Tessellation.SOKOBAN, moves_data=moves_data)


class SokobanPuzzle(Puzzle):
    def __init__(self, width: int = 0, height: int = 0, board: Optional[str] = None):
        """
        Arguments:
            width: number of columns
            height: number of rows
            board: If not blank, it will be parsed and board will be created from it,
                ignoring ``width`` and ``height``.
        """
        from ..game import Tessellation

        super().__init__(
            tessellation=Tessellation.SOKOBAN, width=width, height=height, board=board
        )



import copy

import pytest

from sokoenginepy.game import BoardCell
from sokoenginepy.io import Puzzle


@pytest.fixture
def board_cell():
    return BoardCell()


class DescribeBoardCell:
    class Describe_init:
        def it_creates_empty_floor_by_default(self):
            board_cell = BoardCell()
            assert board_cell.is_empty_floor

        def it_initializes_secondary_flags_to_false_by_default(self):
            board_cell = BoardCell()
            assert not board_cell.is_in_playable_area
            assert not board_cell.is_deadlock

            board_cell: BoardCell = BoardCell(character=Puzzle.PUSHER)
            assert not board_cell.is_in_playable_area
            assert not board_cell.is_deadlock

        def it_raises_on_illegal_character(self):
            with pytest.raises(ValueError):
                BoardCell(character="4")
            with pytest.raises(ValueError):
                BoardCell(character="  ")
            with pytest.raises(ValueError):
                BoardCell(character=None)

    class Describe_equality_comparison:
        def it_doesnt_compare_secondary_flags(self, board_cell):
            board_cell2 = copy.copy(board_cell)
            assert board_cell == board_cell2
            board_cell.is_deadlock = True
            board_cell2.is_deadlock = False
            board_cell.is_in_playable_area = True
            board_cell2.is_in_playable_area = False
            assert board_cell == board_cell2

    class Describe_has_pusher:
        def test_setter_puts_or_removes_pusher_on_cell(self, board_cell):
            board_cell.has_pusher = False
            assert not board_cell.has_pusher
            board_cell.has_pusher = True
            assert board_cell.has_pusher
            board_cell.has_pusher = False
            assert not board_cell.has_pusher

        def test_setter_replaces_box_if_setting_to_true(self, board_cell):
            board_cell.has_box = True
            board_cell.has_pusher = True
            assert not board_cell.has_box

        def test_setter_replaces_box_but_not_goal_if_setting_to_true(self, board_cell):
            board_cell.has_goal = True
            board_cell.has_box = True
            board_cell.has_pusher = True
            assert not board_cell.has_box
            assert board_cell.has_goal

        def test_setter_replaces_wall_if_setting_to_true(self, board_cell):
            board_cell.is_wall = True
            board_cell.has_pusher = True
            assert not board_cell.is_wall
            assert board_cell.has_pusher

        def test_setter_doesnt_change_secondary_flags(self, board_cell):
            board_cell.is_deadlock = True
            board_cell.is_in_playable_area = True
            board_cell.has_pusher = True
            assert board_cell.is_deadlock
            assert board_cell.is_in_playable_area

    class Describe_has_box:
        def test_setter_puts_or_removes_box_on_cell(self, board_cell):
            board_cell.has_box = False
            assert not board_cell.has_box
            board_cell.has_box = True
            assert board_cell.has_box
            board_cell.has_box = False
            assert not board_cell.has_box

        def test_setter_replaces_pusher_if_setting_to_true(self, board_cell):
            board_cell.has_pusher = True
            board_cell.has_box = True
            assert not board_cell.has_pusher

        def test_setter_replaces_pusher_but_not_goal_if_setting_to_true(
            self, board_cell
        ):
            board_cell.has_goal = True
            board_cell.has_pusher = True
            board_cell.has_box = True
            assert not board_cell.has_pusher
            assert board_cell.has_goal

        def test_setter_replaces_wall_if_setting_to_true(self, board_cell):
            board_cell.is_wall = True
            board_cell.has_box = True
            assert not board_cell.is_wall
            assert board_cell.has_box

        def test_setter_doesnt_change_secondary_flags(self, board_cell):
            board_cell.is_deadlock = True
            board_cell.is_in_playable_area = True
            board_cell.has_box = True
            assert board_cell.is_deadlock
            assert board_cell.is_in_playable_area

    class Describe_has_goal:
        def test_setter_puts_or_removes_goal_on_cell(self, board_cell):
            board_cell.has_goal = False
            assert not board_cell.has_goal
            board_cell.has_goal = True
            assert board_cell.has_goal
            board_cell.has_goal = False
            assert not board_cell.has_goal

        def test_setter_adds_or_removes_goal_under_pusher(self, board_cell):
            board_cell.has_pusher = True
            board_cell.has_goal = True
            assert board_cell.has_pusher
            assert board_cell.has_goal
            board_cell.has_goal = False
            assert board_cell.has_pusher
            assert not board_cell.has_goal

        def test_setter_adds_or_removes_goal_under_box(self, board_cell):
            board_cell.has_box = True
            board_cell.has_goal = True
            assert board_cell.has_box
            assert board_cell.has_goal
            board_cell.has_goal = False
            assert board_cell.has_box
            assert not board_cell.has_goal

        def test_setter_replaces_wall_if_setting_to_true(self, board_cell):
            board_cell.is_wall = True
            board_cell.has_goal = True
            assert not board_cell.is_wall
            assert board_cell.has_goal

        def test_setter_doesnt_change_secondary_flags(self, board_cell):
            board_cell.is_deadlock = True
            board_cell.is_in_playable_area = True
            board_cell.has_goal = True
            assert board_cell.is_deadlock
            assert board_cell.is_in_playable_area

    class Describe_is_wall:
        def test_setter_puts_wall_on_cell_if_setting_to_true(self, board_cell):
            board_cell.is_wall = False
            assert not board_cell.is_wall
            board_cell.is_wall = True
            assert board_cell.is_wall
            board_cell.is_wall = False
            assert not board_cell.is_wall

        def test_setter_replaces_box_with_wall_if_setting_to_true(self, board_cell):
            board_cell.has_box = True
            board_cell.is_wall = True
            assert not board_cell.has_box

        def test_setter_replaces_pusher_with_wall_if_setting_to_true(self, board_cell):
            board_cell.has_pusher = True
            board_cell.is_wall = True
            assert not board_cell.has_pusher

        def test_setter_replaces_goal_with_wall_if_setting_to_true(self, board_cell):
            board_cell.has_goal = True
            board_cell.is_wall = True
            assert not board_cell.has_goal

        def test_setter_doesnt_change_secondary_flags(self, board_cell):
            board_cell.is_deadlock = True
            board_cell.is_in_playable_area = True
            board_cell.is_wall = True
            assert board_cell.is_deadlock
            assert board_cell.is_in_playable_area

    class Describe_can_put_pusher_or_box:
        def it_returns_true_for_empty_floor(self, board_cell):
            assert board_cell.can_put_pusher_or_box

        def it_returns_true_for_empty_goal(self, board_cell):
            board_cell.has_goal = True
            assert board_cell.can_put_pusher_or_box

        def it_returns_false_for_wall(self, board_cell):
            board_cell.is_wall = True
            assert not board_cell.can_put_pusher_or_box

        def it_returns_false_for_box(self, board_cell):
            board_cell.has_box = True
            assert not board_cell.can_put_pusher_or_box

        def it_returns_false_for_pusher(self, board_cell):
            board_cell.has_pusher = True
            assert not board_cell.can_put_pusher_or_box

    class Describe_str:
        def it_converts_board_cell_to_printable_character(self, board_cell):
            board_cell.clear()
            board_cell.has_pusher = True
            assert str(board_cell) == Puzzle.PUSHER
            board_cell.clear()
            board_cell.has_pusher = True
            board_cell.has_goal = True
            assert str(board_cell) == Puzzle.PUSHER_ON_GOAL
            board_cell.clear()
            board_cell.has_box = True
            assert str(board_cell) == Puzzle.BOX
            board_cell.clear()
            board_cell.has_box = True
            board_cell.has_goal = True
            assert str(board_cell) == Puzzle.BOX_ON_GOAL
            board_cell.clear()
            board_cell.has_goal = True
            assert str(board_cell) == Puzzle.GOAL
            board_cell.clear()
            board_cell.is_wall = True
            assert str(board_cell) == Puzzle.WALL

            board_cell.clear()

            assert str(board_cell) == Puzzle.FLOOR

            assert board_cell.to_str(use_visible_floor=True) == Puzzle.VISIBLE_FLOOR



from sokoenginepy.game import BaseTessellation, Direction, PusherStep, Tessellation

from .autogenerated_tessellation import HexobanTessellationAutogeneratedSpecMixin
from .tessellation_spec_mixin import TessellationSpecMixin


class DescribeHexobanTessellation(
    TessellationSpecMixin, HexobanTessellationAutogeneratedSpecMixin
):

    illegal_directions = [Direction.UP, Direction.DOWN]

    legal_directions = [
        Direction.LEFT,
        Direction.RIGHT,
        Direction.NORTH_WEST,
        Direction.NORTH_EAST,
        Direction.SOUTH_WEST,
        Direction.SOUTH_EAST,
    ]

    legal_moves = [
        PusherStep(Direction.LEFT, False),
        PusherStep(Direction.RIGHT, False),
        PusherStep(Direction.NORTH_WEST, False),
        PusherStep(Direction.NORTH_EAST, False),
        PusherStep(Direction.SOUTH_WEST, False),
        PusherStep(Direction.SOUTH_EAST, False),
        PusherStep(Direction.LEFT, True),
        PusherStep(Direction.RIGHT, True),
        PusherStep(Direction.NORTH_WEST, True),
        PusherStep(Direction.NORTH_EAST, True),
        PusherStep(Direction.SOUTH_WEST, True),
        PusherStep(Direction.SOUTH_EAST, True),
    ]

    legal_characters = ["l", "r", "u", "n", "s", "d", "L", "R", "U", "N", "S", "D"]

    tessellation = BaseTessellation.instance(Tessellation.HEXOBAN)



import textwrap

import pytest

from sokoenginepy.game import Tessellation, index_1d
from sokoenginepy.io import Puzzle, SokobanPuzzle

from ..fixtures import BoardData


@pytest.fixture
def input_data():
    data = """
            #####
            #   #
            #$  #
          ###  $##
          #  $ $ #
        ### # ## #   ######
        #   # ## #####  ..#
        # $  $          ..#
        ##### ### #@##  ..#
            #     #########
            #######
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return BoardData(board=data, width=19, height=11)


@pytest.fixture
def parsed_data():
    data = """
        ----#####----------
        ----#---#----------
        ----#$--#----------
        --###--$##---------
        --#--$-$-#---------
        ###-#-##-#---######
        #---#-##-#####--..#
        #-$--$----------..#
        #####-###-#@##--..#
        ----#-----#########
        ----#######--------
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return BoardData(board=data, width=19, height=11)


class DescribeSokobanPuzzle:
    def it_creates_board_of_specified_size_and_tessellation(self):
        b = SokobanPuzzle(4, 2)
        assert b.width == 4
        assert b.height == 2
        assert b.tessellation == Tessellation.SOKOBAN

    def it_ignores_specified_size_if_board_string_is_given_and_parses_given_string_instead(
        self,
    ):
        puzzle = SokobanPuzzle(4, 2, board="#        #")
        assert puzzle.width == 10
        assert puzzle.height == 1
        assert puzzle.board == "#        #"
        assert str(puzzle) == "#--------#"

    def it_carries_some_metadata(self):
        puzzle = SokobanPuzzle()
        puzzle.title = "title"
        assert puzzle.title == "title"
        puzzle.author = "author"
        assert puzzle.author == "author"
        puzzle.boxorder = "boxorder"
        assert puzzle.boxorder == "boxorder"
        puzzle.goalorder = "goalorder"
        assert puzzle.goalorder == "goalorder"

    def it_can_access_individual_board_characters(self):
        puzzle = SokobanPuzzle(width=5, height=5)
        puzzle[index_1d(3, 3, puzzle.width)] = Puzzle.WALL
        assert puzzle[index_1d(3, 3, puzzle.width)] == Puzzle.WALL

    def it_validates_input_characters_when_setting_them(self):
        puzzle = SokobanPuzzle(width=5, height=5)
        with pytest.raises(ValueError):
            puzzle[index_1d(3, 3, puzzle.width)] = "Z"

    def it_provides_underlying_tessellation(self):
        puzzle = SokobanPuzzle()
        assert puzzle.tessellation == Tessellation.SOKOBAN

    def it_provides_board_dimensions(self):
        puzzle = SokobanPuzzle(4, 2)
        assert puzzle.width == 4
        assert puzzle.height == 2
        assert puzzle.size == 8

    def it_provides_agents_counts(self):
        puzzle = SokobanPuzzle(board="#@@...$$$$#")
        assert puzzle.pushers_count == 2
        assert puzzle.goals_count == 3
        assert puzzle.boxes_count == 4

    def it_provides_internal_and_original_board_strings(self):
        puzzle = SokobanPuzzle(4, 2)
        assert puzzle.internal_board == "--------"
        assert puzzle.board == ""
        assert puzzle.width == 4
        assert puzzle.height == 2

        puzzle = SokobanPuzzle(board="#       #")
        assert puzzle.internal_board == "#-------#"
        assert puzzle.board == "#       #"
        assert puzzle.width == 9
        assert puzzle.height == 1

    def it_provides_conversions_to_string(self):
        puzzle = SokobanPuzzle(board="#       #")

        assert puzzle.board == "#       #"
        assert str(puzzle) == "#-------#"
        assert puzzle.to_board_str(use_visible_floor=False) == "#       #"
        assert puzzle.to_board_str(use_visible_floor=True) == "#-------#"

    class describe_parsing_from_string:
        def it_parses_board_from_string(
            self, input_data: BoardData, parsed_data: BoardData
        ):
            puzzle = SokobanPuzzle(board=input_data.board)
            assert str(puzzle) == parsed_data.board
            assert puzzle.width == parsed_data.width
            assert puzzle.height == parsed_data.height

        def it_raises_on_illegal_characters_in_board_string(self):
            with pytest.raises(ValueError):
                SokobanPuzzle(board="ZOMG!")

            with pytest.raises(ValueError):
                puzzle = SokobanPuzzle()
                puzzle.board = "ZOMG!"

        def it_correctly_parses_board_with_blank_rows(self):
            src = (
                " ##########\n    \n"
                " ## @ *   #\n"
                "##  $ ####\n"
                "# $.+ #\n"
                "#######\n"
            )
            result = "\n".join(
                [
                    "-##########",
                    "-----------",
                    "-##-@-*---#",
                    "##--$-####-",
                    "#-$.+-#----",
                    "#######----",
                    "-----------",
                ]
            )
            puzzle = SokobanPuzzle(board=src)
            assert str(puzzle) == result

        def it_parses_RLE_encoded_board(self):
            src = (
                " ##########   |"
                " ## @ *   #|"
                "##  $ ####|"
                "# $.+ #|||"
                "#######   |"
            )
            result = "\n".join(
                [
                    "-##########---",
                    "-##-@-*---#---",
                    "##--$-####----",
                    "#-$.+-#-------",
                    "--------------",
                    "--------------",
                    "#######-------",
                    "--------------",
                ]
            )
            puzzle = SokobanPuzzle(board=src)
            assert str(puzzle) == result

        def it_carefully_preserves_original_newlines(self):
            src = (
                " ##########\n    \n\n"
                " ## @ *   #\n\n"
                "##  $ ####\n\n"
                "# $.+ #\n\n"
                "#######\n\n"
            )
            result = "\n".join(
                [
                    "-##########",
                    "-----------",
                    "-----------",
                    "-##-@-*---#",
                    "-----------",
                    "##--$-####-",
                    "-----------",
                    "#-$.+-#----",
                    "-----------",
                    "#######----",
                    "-----------",
                    "-----------",
                ]
            )
            puzzle = SokobanPuzzle(board=src)
            assert str(puzzle) == result

    class describe_resize:
        def it_adds_right_columns_and_bottom_rows_when_enlarging(
            self, input_data: BoardData
        ):
            data = """
                ----#####------------
                ----#---#------------
                ----#$--#------------
                --###--$##-----------
                --#--$-$-#-----------
                ###-#-##-#---######--
                #---#-##-#####--..#--
                #-$--$----------..#--
                #####-###-#@##--..#--
                ----#-----#########--
                ----#######----------
                ---------------------
                ---------------------
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.resize(puzzle.width + 2, puzzle.height + 2)
            assert puzzle.width == 21
            assert puzzle.height == 13
            assert str(puzzle) == data

        def it_removes_right_columns_and_bottom_rows_when_shrinking(
            self, input_data: BoardData
        ):
            data = """
                ----#####--------
                ----#---#--------
                ----#$--#--------
                --###--$##-------
                --#--$-$-#-------
                ###-#-##-#---####
                #---#-##-#####--.
                #-$--$----------.
                #####-###-#@##--.
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.resize(puzzle.width - 2, puzzle.height - 2)
            assert puzzle.width == 17
            assert puzzle.height == 9
            assert str(puzzle) == data

    class describe_resize_and_center:
        def it_adds_columns_and_rows_when_enlarging(self, input_data: BoardData):
            data = """
                ------------------------
                ------------------------
                ------#####-------------
                ------#---#-------------
                ------#$--#-------------
                ----###--$##------------
                ----#--$-$-#------------
                --###-#-##-#---######---
                --#---#-##-#####--..#---
                --#-$--$----------..#---
                --#####-###-#@##--..#---
                ------#-----#########---
                ------#######-----------
                ------------------------
                ------------------------
                ------------------------
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.resize_and_center(puzzle.width + 5, puzzle.height + 5)
            assert puzzle.width == 24
            assert puzzle.height == 16
            assert str(puzzle) == data

        def it_doesnt_remove_columns_or_rows_when_compacting(
            self, input_data: BoardData, parsed_data: BoardData
        ):
            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.resize_and_center(puzzle.width - 2, puzzle.height - 2)
            assert puzzle.width == parsed_data.width
            assert puzzle.height == parsed_data.height
            assert str(puzzle) == parsed_data.board

    class describe_trim:
        def it_removes_empty_outer_rows_and_columns(self, parsed_data: BoardData):
            data = """
                ------------------------
                ------------------------
                ------#####-------------
                ------#---#-------------
                ------#$--#-------------
                ----###--$##------------
                ----#--$-$-#------------
                --###-#-##-#---######---
                --#---#-##-#####--..#---
                --#-$--$----------..#---
                --#####-###-#@##--..#---
                ------#-----#########---
                ------#######-----------
                ------------------------
                ------------------------
                ------------------------
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=data)
            puzzle.trim()
            assert puzzle.width == parsed_data.width
            assert puzzle.height == parsed_data.height
            assert str(puzzle) == parsed_data.board

    class describe_reverse_rows:
        def it_mirrors_board_up_down(self, input_data: BoardData):
            data = """
                ----#######--------
                ----#-----#########
                #####-###-#@##--..#
                #-$--$----------..#
                #---#-##-#####--..#
                ###-#-##-#---######
                --#--$-$-#---------
                --###--$##---------
                ----#$--#----------
                ----#---#----------
                ----#####----------
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.reverse_rows()
            assert puzzle.width == input_data.width
            assert puzzle.height == input_data.height
            assert str(puzzle) == data

    class describe_reverse_columns:
        def it_mirrors_board_left_right(self, input_data: BoardData):
            data = """
                ----------#####----
                ----------#---#----
                ----------#--$#----
                ---------##$--###--
                ---------#-$-$--#--
                ######---#-##-#-###
                #..--#####-##-#---#
                #..----------$--$-#
                #..--##@#-###-#####
                #########-----#----
                --------#######----
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())

            puzzle = SokobanPuzzle(board=input_data.board)
            puzzle.reverse_columns()
            assert puzzle.width == input_data.width
            assert puzzle.height == input_data.height
            assert str(puzzle) == data



from ..io import Puzzle


class BoardCell:
    """
    Stores properties of one cell in board layout.

    Note:
        There is no game logic encoded in this class. It is perfectly fine to put
        pusher on wall cell (in which case wall will be replaced by pusher). This is
        by design: :class:`.BoardCell` is value class, not game logic class.
    """

    __slots__ = [
        "_has_box",
        "_has_pusher",
        "_has_goal",
        "_is_wall",
        "is_in_playable_area",
        "is_deadlock",
    ]

    def __init__(
        self,
        character: str = Puzzle.FLOOR,
        is_in_playable_area: bool = False,
        is_deadlock: bool = False,
    ):
        self._has_box: bool = False
        self._has_pusher: bool = False
        self._has_goal: bool = False
        self._is_wall: bool = False
        self.is_in_playable_area: bool = is_in_playable_area
        self.is_deadlock: bool = is_deadlock

        if not Puzzle.is_empty_floor(character):
            if Puzzle.is_wall(character):
                self.is_wall = True
            elif Puzzle.is_pusher(character):
                self.has_pusher = True
                if Puzzle.is_goal(character):
                    self.has_goal = True
            elif Puzzle.is_box(character):
                self.has_box = True
                if Puzzle.is_goal(character):
                    self.has_goal = True
            elif Puzzle.is_goal(character):
                self.has_goal = True
            else:
                raise ValueError("Illegal characters found in board string")

    def __eq__(self, rv):
        return (
            isinstance(rv, self.__class__)
            and self.is_wall == rv.is_wall
            and self.has_pusher == rv.has_pusher
            and self.has_box == rv.has_box
            and self.has_goal == rv.has_goal
        ) or (
            isinstance(rv, str)
            and self.is_wall == Puzzle.is_wall(rv)
            and self.has_pusher == Puzzle.is_pusher(rv)
            and self.has_box == Puzzle.is_box(rv)
            and self.has_goal == Puzzle.is_goal(rv)
        )

    def __ne__(self, rv):
        return not self == rv

    def __str__(self):
        return self.to_str(use_visible_floor=False)

    def __repr__(self):
        return "BoardCell('{0}')".format(str(self))

    def to_str(self, use_visible_floor: bool = False) -> str:
        retv = Puzzle.FLOOR

        if not self.has_piece:
            if self.is_wall:
                retv = Puzzle.WALL
            else:
                retv = Puzzle.VISIBLE_FLOOR if use_visible_floor else Puzzle.FLOOR
        elif not self.has_box and not self.has_goal and self.has_pusher:
            retv = Puzzle.PUSHER
        elif not self.has_box and self.has_goal and not self.has_pusher:
            retv = Puzzle.GOAL
        elif not self.has_box and self.has_goal and self.has_pusher:
            retv = Puzzle.PUSHER_ON_GOAL
        elif self.has_box and not self.has_goal and not self.has_pusher:
            retv = Puzzle.BOX
        else:
            retv = Puzzle.BOX_ON_GOAL

        return retv

    def clear(self):
        """Clears cell, converting it to empty floor."""
        self._is_wall = self._has_box = self._has_goal = self._has_pusher = False

    @property
    def has_piece(self) -> bool:
        """True if there is pusher, box or goal on this cell."""
        return self.has_pusher or self.has_box or self.has_goal

    @property
    def is_empty_floor(self) -> bool:
        """True if there is no pieces and no wall on this cell."""
        return not (self.has_pusher or self.has_box or self.has_goal or self.is_wall)

    @property
    def is_border_element(self) -> bool:
        """True if this is either a wall or box on goal."""
        return self.is_wall or (self.has_box and self.has_goal)

    @property
    def can_put_pusher_or_box(self) -> bool:
        """
        True if this cell allows putting box or pusher on self.

        Note:
            This method is not used by BoardCell modifiers (ie. `put_box`,
            `put_pusher`, etc...). As far as BoardCell is concerned, nothing prevents
            clients from putting box on wall (which replaces that wall with box).
            This method is intended to be used by higher game logic classes that
            would implement pusher movement in which case putting ie. pusher onto
            same cell where box is makes no sense.
        """
        return not (self.has_box or self.has_pusher or self.is_wall)

    @property
    def has_box(self) -> bool:
        return self._has_box

    @has_box.setter
    def has_box(self, value: bool):
        if value:
            self._has_box = True
            self._is_wall = False
            self._has_pusher = False
        else:
            self._has_box = False

    def put_box(self):
        self.has_box = True

    def remove_box(self):
        self.has_box = False

    @property
    def has_goal(self) -> bool:
        return self._has_goal

    @has_goal.setter
    def has_goal(self, value: bool):
        if value:
            self._has_goal = True
            self._is_wall = False
        else:
            self._has_goal = False

    def put_goal(self):
        self.has_goal = True

    def remove_goal(self):
        self.has_goal = False

    @property
    def has_pusher(self):
        return self._has_pusher

    @has_pusher.setter
    def has_pusher(self, value: bool):
        if value:
            self._has_pusher = True
            self._has_box = False
            self._is_wall = False
        else:
            self._has_pusher = False

    def put_pusher(self):
        self.has_pusher = True

    def remove_pusher(self):
        self.has_pusher = False

    @property
    def is_wall(self) -> bool:
        return self._is_wall

    @is_wall.setter
    def is_wall(self, value: bool):
        if value:
            self._is_wall = True
            self._has_box = False
            self._has_goal = False
            self._has_pusher = False
        else:
            self._is_wall = False



from sokoenginepy.game import X, Y, index_1d, is_on_board_1d, is_on_board_2d


class Describe_INDEX:
    def it_calculates_1D_index_from_2D_coordinates(self):
        assert index_1d(0, 0, 5) == 0
        assert index_1d(4, 5, 5) == 29
        assert index_1d(3, 3, 5) == 18


class Describe_X:
    def it_calculates_X_coordinate_from_1D_index(self):
        assert X(0, 5) == 0
        assert X(29, 5) == 4
        assert X(18, 5) == 3


class Describe_Y:
    def it_calculates_Y_coordinate_from_1D_index(self):
        assert Y(0, 5) == 0
        assert Y(29, 5) == 5
        assert Y(18, 5) == 3


class Describe_on_board_2d:
    def it_checks_if_2D_coordinate_is_on_board(self):
        assert is_on_board_2d(0, 0, 3, 3)
        assert is_on_board_2d(1, 1, 3, 3)
        assert is_on_board_2d(2, 2, 3, 3)
        assert not is_on_board_2d(3, 3, 3, 3)


class Describe_on_board_1d:
    def it_checks_if_1D_coordinate_is_on_board(self):
        assert is_on_board_1d(0, 3, 3)
        assert is_on_board_1d(4, 3, 3)
        assert is_on_board_1d(8, 3, 3)
        assert not is_on_board_1d(9, 3, 3)



from __future__ import annotations

import operator
import textwrap
import time
from enum import IntEnum
from functools import reduce

from sokoenginepy.game import BoardGraph, Direction, Mover, SolvingMode
from sokoenginepy.io import SokobanPuzzle


class BoardType(IntEnum):
    SMALL = 1
    LARGE = 2

    @property
    def puzzle(self) -> SokobanPuzzle:
        if self == self.SMALL:
            data = """
                ##########
                #      **#
                #      **#
                # *@   **#
                #      **#
                ##########
            """
            data = textwrap.dedent(data.lstrip("\n").rstrip())
            return SokobanPuzzle(board=data)
        else:
            data = """
                ######################################
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #*************          *************#
                #*************          *************#
                #*************          *************#
                #*************    *@    *************#
                #*************          *************#
                #*************          *************#
                #*************          *************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                #************************************#
                ######################################
            """
            return SokobanPuzzle(board=data)


class BenchmarkType(IntEnum):
    FORWARD_MOVER = 0
    REVERSE_MOVER = 1

    @property
    def is_reverse(self) -> bool:
        return self == self.REVERSE_MOVER

    @property
    def title(self) -> str:
        return self.name.replace("_", " ").lower().title()

    @property
    def direction(self) -> Direction:
        if self == self.FORWARD_MOVER:
            return Direction.LEFT
        else:
            return Direction.RIGHT


class MovementBenchmark:
    def __init__(
        self, board_type: BoardType, benchmark_type: BenchmarkType, moves_count: int
    ):
        self.milliseconds_used = 0
        self.board_type = board_type
        self.benchmark_type = benchmark_type
        self.moves_count = moves_count
        self.puzzle = self.board_type.puzzle
        self.graph = BoardGraph(self.puzzle)

        self.mover = Mover(
            self.graph,
            (
                SolvingMode.REVERSE
                if self.benchmark_type.is_reverse
                else SolvingMode.FORWARD
            ),
        )
        self.mover.pulls_boxes = True

    @property
    def moves_per_second(self) -> float:
        return self.moves_count / (self.milliseconds_used / 1000)

    def run(self):
        total_time = 0
        undo_move = False
        for i in range(0, int(self.moves_count)):
            if undo_move:
                start_time = time.perf_counter()
                self.mover.undo_last_move()
                end_time = time.perf_counter()
                total_time += end_time - start_time
                undo_move = False
            else:
                start_time = time.perf_counter()
                self.mover.move(self.benchmark_type.direction)
                end_time = time.perf_counter()
                total_time += end_time - start_time
                undo_move = True

        self.milliseconds_used = total_time * 1000


class MovementBenchmarkPrinter:
    def __init__(self, runs_count: int, moves_per_run_count: int):
        self.runs_count = runs_count
        self.moves_per_run_count = moves_per_run_count

    def board_header(self, board_type):
        benchmarker = MovementBenchmark(board_type, BenchmarkType.FORWARD_MOVER, 1)

        return (
            "{:<10} W:{:<5} H:{:<5} P:{:<5} B:{:<5}".format(
                "Board: ",
                benchmarker.puzzle.width,
                benchmarker.puzzle.height,
                benchmarker.mover.board_manager.pushers_count,
                benchmarker.mover.board_manager.boxes_count,
            )
            + "\n"
            + str(benchmarker.mover.board)
            + "\nMoves count: {0}".format(self.moves_per_run_count)
        )

    def run_and_print_experiment(
        self, board_type: BoardType, benchmark_type: BenchmarkType, pivot_speed=None
    ):
        speeds = []
        times = []

        print("{:<20}: ".format(benchmark_type.title), end="", flush=True)

        for i in range(0, self.runs_count):
            benchmarker = MovementBenchmark(
                board_type, benchmark_type, self.moves_per_run_count
            )
            benchmarker.run()
            speeds.append(benchmarker.moves_per_second)
            times.append(benchmarker.milliseconds_used)
            print(".", end="", flush=True)

        mean_speed = reduce(operator.add, speeds, 0.0) / len(speeds)
        mean_time = reduce(operator.add, times, 0.0) / len(times)
        print(
            " {:.2f} [ms] {:.2e} [moves/s]".format(mean_time, mean_speed),
            end="",
            flush=True,
        )

        if pivot_speed:
            print("  {0:.2f}%".format(mean_speed / pivot_speed * 100))
        else:
            print("  100.00%")

        return mean_speed

    @classmethod
    def run_all(cls):
        print("--------------------------------------------------")
        print("--              MOVER BENCHMARKS                --")
        print("--------------------------------------------------")

        native_extension_available = True
        try:
            import sokoenginepyext
        except ImportError:
            native_extension_available = False

        if native_extension_available:
            runs = 10
            moves_per_run = 3e5
        else:
            runs = 10
            moves_per_run = 3e4

        printer = MovementBenchmarkPrinter(runs, int(moves_per_run))
        print(printer.board_header(BoardType.SMALL))

        # Speed of early, not thoroughly tested, C++ only implementation
        pivot_speed = 3e6

        printer.run_and_print_experiment(
            BoardType.SMALL, BenchmarkType.FORWARD_MOVER, pivot_speed
        )
        printer.run_and_print_experiment(
            BoardType.SMALL, BenchmarkType.REVERSE_MOVER, pivot_speed
        )


def run_benchmarks():
    return MovementBenchmarkPrinter.run_all()



from sokoenginepy.io import Puzzle


class DescribePuzzle:
    def it_correctly_categorizes_pusher_characters(self):
        assert Puzzle.is_pusher(Puzzle.PUSHER)
        assert Puzzle.is_pusher(Puzzle.ALT_PUSHER1)
        assert Puzzle.is_pusher(Puzzle.ALT_PUSHER2)
        assert Puzzle.is_pusher(Puzzle.PUSHER_ON_GOAL)
        assert Puzzle.is_pusher(Puzzle.ALT_PUSHER_ON_GOAL1)
        assert Puzzle.is_pusher(Puzzle.ALT_PUSHER_ON_GOAL2)

    def it_correctly_categorizes_box_characters(self):
        assert Puzzle.is_box(Puzzle.BOX)
        assert Puzzle.is_box(Puzzle.ALT_BOX1)
        assert Puzzle.is_box(Puzzle.BOX_ON_GOAL)
        assert Puzzle.is_box(Puzzle.ALT_BOX_ON_GOAL1)

    def it_correctly_categorizes_goal_characters(self):
        assert Puzzle.is_goal(Puzzle.GOAL)
        assert Puzzle.is_goal(Puzzle.ALT_GOAL1)
        assert Puzzle.is_goal(Puzzle.PUSHER_ON_GOAL)
        assert Puzzle.is_goal(Puzzle.ALT_PUSHER_ON_GOAL1)
        assert Puzzle.is_goal(Puzzle.ALT_PUSHER_ON_GOAL2)
        assert Puzzle.is_goal(Puzzle.BOX_ON_GOAL)
        assert Puzzle.is_goal(Puzzle.ALT_BOX_ON_GOAL1)

    def it_correctly_categorizes_empty_floor_characters(self):
        assert Puzzle.is_empty_floor(Puzzle.FLOOR)
        assert Puzzle.is_empty_floor(Puzzle.VISIBLE_FLOOR)
        assert Puzzle.is_empty_floor(Puzzle.ALT_VISIBLE_FLOOR1)

    def it_correctly_categorizes_wall_characters(self):
        assert Puzzle.is_wall(Puzzle.WALL)

    def it_correctly_categorizes_border_characters(self):
        assert Puzzle.is_border_element(Puzzle.WALL)
        assert Puzzle.is_border_element(Puzzle.BOX_ON_GOAL)
        assert Puzzle.is_border_element(Puzzle.ALT_BOX_ON_GOAL1)

    def it_can_categorize_any_character(self):
        assert Puzzle.is_puzzle_element(Puzzle.WALL)
        assert not Puzzle.is_puzzle_element("Z")

    class describe_is_board:
        def it_recognizes_board_string(self):
            assert Puzzle.is_board("0123456789\n\t bB$*p_pmM@+#_-|")

        def it_fails_on_illegal_characters(self):
            assert not Puzzle.is_board("ZOMG!")

        def it_fails_on_purely_numeric_strings(self):
            assert not Puzzle.is_board("42")

        def it_fails_on_blank_strings(self):
            assert not Puzzle.is_board("")
            assert not Puzzle.is_board("    ")
            assert not Puzzle.is_board("   \r\n \t")

    class describe_is_sokoban_plus:
        def it_recognizes_sokoban_plus_strings(self):
            assert Puzzle.is_sokoban_plus("0 1 2 3 4 5 6 99")

        def it_fails_on_blank_strings(self):
            assert not Puzzle.is_sokoban_plus("")
            assert not Puzzle.is_sokoban_plus("    ")
            assert not Puzzle.is_sokoban_plus("   \r\n \t")



import textwrap
from itertools import permutations

import pytest

from sokoenginepy.game import BoardCell, BoardGraph, Direction, index_1d
from sokoenginepy.io import Puzzle, SokobanPuzzle, TriobanPuzzle


@pytest.fixture
def puzzle():
    data = """
            #####
            #   #
            #$  #
          ###  $##
          #  $ $ #
        ### # ## #   ######
        #   # ## #####  ..#
        # $  $          ..#
        ##### ### #@##  ..#
            #     #########
            #######
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return SokobanPuzzle(board=data)


@pytest.fixture
def solved_puzzle():
    data = "\n".join(
        [
            # 123456789012345678
            "    #####          ",  # 0
            "    #  @#          ",  # 1
            "    #   #          ",  # 2
            "  ###   ##         ",  # 3
            "  #      #         ",  # 4
            "### # ## #   ######",  # 5
            "#   # ## #####  **#",  # 6
            "#               **#",  # 7
            "##### ### #@##  **#",  # 8
            "    #     #########",  # 9
            "    #######        ",  # 0
        ]
    )
    return SokobanPuzzle(board=data)


@pytest.fixture
def board_graph(puzzle):
    return BoardGraph(puzzle)


@pytest.fixture
def positions_path(puzzle):
    return [
        index_1d(7, 1, puzzle.width),
        index_1d(6, 1, puzzle.width),
        index_1d(6, 2, puzzle.width),
        index_1d(6, 3, puzzle.width),
        index_1d(6, 4, puzzle.width),
        index_1d(5, 4, puzzle.width),
    ]


@pytest.fixture
def directions_path():
    return [
        Direction.LEFT,
        Direction.DOWN,
        Direction.DOWN,
        Direction.DOWN,
        Direction.LEFT,
    ]


class DescribeBoardGraph:
    class describe_getitem:
        def it_returns_board_cell_on_position(self, board_graph):
            assert isinstance(board_graph[0], BoardCell)

        def it_raises_if_trying_to_get_invalid_position(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph[42000]
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph[_]

    class describe_setitem:
        def it_sets_board_cell_on_position(self, board_graph):
            board_graph[0].is_in_playable_area = True
            board_graph[0].is_deadlock = True
            board_graph[0].is_wall = True

            bc = BoardCell(Puzzle.PUSHER)
            bc.is_in_playable_area = False
            bc.is_deadlock = False

            board_graph[0] = bc

            assert not board_graph[0].is_wall
            assert board_graph[0].has_pusher
            assert not board_graph[0].is_in_playable_area
            assert not board_graph[0].is_deadlock

            board_graph[0] = Puzzle.PUSHER

            assert not board_graph[0].is_wall
            assert board_graph[0].has_pusher
            assert not board_graph[0].is_in_playable_area
            assert not board_graph[0].is_deadlock

        def it_raises_if_trying_to_set_invalid_cell_value(self, board_graph):
            with pytest.raises(ValueError):
                board_graph[0] = ""
            with pytest.raises(ValueError):
                board_graph[0] = "  "
            with pytest.raises(ValueError):
                board_graph[0] = None

        def it_raises_if_trying_to_set_invalid_position(
            self, is_using_native, board_graph
        ):
            board_cell = BoardCell()
            with pytest.raises(IndexError):
                board_graph[42000] = board_cell
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph[_] = board_cell

    class describe_contains:
        def it_detects_if_position_is_in_graph(self, board_graph):
            assert 0 in board_graph

        def it_doesnt_rise_if_checking_invalid_position_value(self, board_graph):
            for _ in [42000, -42000, None, "", "ZOMG!", "42"]:
                assert _ not in board_graph

    class describe_vertices_count:
        def it_returns_number_of_graph_positions(self, puzzle, board_graph):
            assert board_graph.vertices_count == puzzle.width * puzzle.height

    class describe_edges_count:
        def it_returns_number_of_graph_edges(self, board_graph):
            assert board_graph.edges_count == 776

    class describe_has_edge:
        def it_checks_if_edge_in_given_direction_exists(self, board_graph):
            assert board_graph.has_edge(0, 1, Direction.RIGHT)
            assert not board_graph.has_edge(0, 1, Direction.LEFT)

        def it_doesnt_rise_if_checking_invalid_position_value(self, board_graph):
            for _ in [42000, -42000, None, "", "ZOMG!", "42"]:
                assert not board_graph.has_edge(_, 1, Direction.LEFT)
                assert not board_graph.has_edge(1, _, Direction.LEFT)

    class describe_out_edges_count:
        def it_returns_number_of_edges_going_from_source_to_target_position(
            self, board_graph
        ):
            assert board_graph.out_edges_count(0, 1) == 1

        def it_doesnt_rise_if_checking_invalid_position_value(self, board_graph):
            for _ in [42000, -42000, None, "", "ZOMG!", "42"]:
                assert board_graph.out_edges_count(_, 1) == 0
                assert board_graph.out_edges_count(1, _) == 0

    class describe_remove_all_edges:
        def it_removes_all_edges_from_graph(self, board_graph):
            board_graph.remove_all_edges()
            assert board_graph.edges_count == 0

    class describe_add_edge:
        def it_adds_edge_between_two_positions(self, board_graph):
            board_graph.remove_all_edges()
            board_graph.add_edge(0, 1, Direction.LEFT)
            assert board_graph.edges_count == 1
            assert board_graph.has_edge(0, 1, Direction.LEFT)

        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            board_graph.remove_all_edges()

            with pytest.raises(IndexError):
                board_graph.add_edge(42000, 1, Direction.LEFT)
            with pytest.raises(IndexError):
                board_graph.add_edge(1, 42000, Direction.LEFT)

            exc_cls = TypeError if is_using_native else KeyError
            for k in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.add_edge(k, 1, Direction.LEFT)
                with pytest.raises(exc_cls):
                    board_graph.add_edge(1, k, Direction.LEFT)

        def it_allows_adding_duplicate_edges(self):
            board_graph = BoardGraph(SokobanPuzzle(width=2, height=2))
            board_graph.remove_all_edges()
            board_graph.add_edge(0, 1, Direction.LEFT)
            board_graph.add_edge(0, 1, Direction.LEFT)
            assert board_graph.edges_count == 1
            assert board_graph.has_edge(0, 1, Direction.LEFT)

            board_graph = BoardGraph(TriobanPuzzle(width=2, height=2))
            board_graph.remove_all_edges()
            board_graph.add_edge(0, 1, Direction.LEFT)
            board_graph.add_edge(0, 1, Direction.LEFT)
            assert board_graph.edges_count == 2
            assert board_graph.has_edge(0, 1, Direction.LEFT)

    class describe_out_edge_weight:
        def it_returns_max_weight_for_wall_cell_target(
            self, is_using_native, board_graph
        ):
            board_graph[1].is_wall = True
            if not is_using_native:
                assert board_graph.out_edge_weight(1) > len(Direction)
            else:
                assert board_graph.out_edge_weight(1) > Direction.__len__()

        def it_returns_max_weight_for_pusher_cell_target(
            self, is_using_native, board_graph
        ):
            board_graph[1].has_pusher = True
            if not is_using_native:
                assert board_graph.out_edge_weight(1) > len(Direction)
            else:
                assert board_graph.out_edge_weight(1) > Direction.__len__()

        def it_returns_max_weight_for_box_cell_target(
            self, is_using_native, board_graph
        ):
            board_graph[1].has_box = True
            if not is_using_native:
                assert board_graph.out_edge_weight(1) > len(Direction)
            else:
                assert board_graph.out_edge_weight(1) > Direction.__len__()

        def it_returns_one_for_other_cells(self, board_graph):
            board_graph[1].clear()
            assert board_graph.out_edge_weight(1) == 1
            board_graph[1].has_goal = True
            assert board_graph.out_edge_weight(1) == 1

        def it_raises_if_position_is_invalid_value(self, is_using_native, board_graph):
            with pytest.raises(IndexError):
                board_graph.out_edge_weight(42000)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.out_edge_weight(_)

    class describe_neighbor:
        def it_returns_neighbor_position_in_given_direction(self, board_graph):
            assert board_graph.neighbor(0, Direction.RIGHT) == 1

        def it_returns_none_for_off_board_direction(self, board_graph):
            assert board_graph.neighbor(0, Direction.UP) is None

        def it_raises_if_position_is_invalid_value(self, is_using_native, board_graph):
            with pytest.raises(IndexError):
                board_graph.neighbor(42000, Direction.UP)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.neighbor(_, Direction.UP)

    class describe_wall_neighbors:
        def it_returns_positions_of_walls_for_given_vertice(self):
            board_str = "\n".join(
                [
                    # 123456
                    "#######",  # 0
                    "#.$# @#",  # 1
                    "#######",  # 2
                ]
            )
            board_graph = BoardGraph(SokobanPuzzle(board=board_str))
            wall_neighbors = board_graph.wall_neighbors(0)
            assert index_1d(0, 1, 7) in wall_neighbors
            assert index_1d(1, 0, 7) in wall_neighbors

        def it_raises_if_position_is_invalid_value(self, is_using_native, board_graph):
            with pytest.raises(IndexError):
                board_graph.wall_neighbors(42000)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.wall_neighbors(_)

    class describe_all_neighbors:
        def it_returns_positions_of_all_neghbor_positions_for_given_vertice(
            self, board_graph
        ):
            all_neighbors = board_graph.all_neighbors(0)
            assert index_1d(0, 1, board_graph.board_width) in all_neighbors
            assert index_1d(1, 1, board_graph.board_width) not in all_neighbors
            assert index_1d(1, 0, board_graph.board_width) in all_neighbors

        def it_raises_if_position_is_invalid_value(self, is_using_native, board_graph):
            with pytest.raises(IndexError):
                board_graph.all_neighbors(42000)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.all_neighbors(_)

    class describe_shortest_path:
        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.shortest_path(1, 42000)
            with pytest.raises(IndexError):
                board_graph.shortest_path(42000, 1)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.shortest_path(_, 1)
                with pytest.raises(exc_cls):
                    board_graph.shortest_path(1, _)

    class describe_dijkstra_path:
        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.dijkstra_path(1, 42000)
            with pytest.raises(IndexError):
                board_graph.dijkstra_path(42000, 1)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.dijkstra_path(_, 1)
                with pytest.raises(exc_cls):
                    board_graph.dijkstra_path(1, _)

    class describe_find_jump_path:
        def it_returns_sequence_of_positions_defining_shortest_path_for_pusher_jump(
            self, board_graph
        ):
            start_position = index_1d(11, 8, board_graph.board_width)
            end_position = index_1d(8, 5, board_graph.board_width)
            expected = board_graph.positions_path_to_directions_path(
                board_graph.find_jump_path(start_position, end_position)
            )

            assert tuple(expected) in permutations(
                [
                    Direction.UP,
                    Direction.UP,
                    Direction.UP,
                    Direction.LEFT,
                    Direction.LEFT,
                    Direction.LEFT,
                ]
            )

        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.find_jump_path(1, 42000)
            with pytest.raises(IndexError):
                board_graph.find_jump_path(42000, 1)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.find_jump_path(_, 1)
                with pytest.raises(exc_cls):
                    board_graph.find_jump_path(1, _)

    class describe_find_move_path:
        def it_returns_sequence_of_positions_defining_shortest_path_for_pusher_movement_without_pushing_boxes(
            self, board_graph
        ):
            start_position = index_1d(11, 8, board_graph.board_width)
            end_position = index_1d(8, 5, board_graph.board_width)
            expected = board_graph.positions_path_to_directions_path(
                board_graph.find_move_path(start_position, end_position)
            )

            assert tuple(expected) in permutations(
                [
                    Direction.LEFT,
                    Direction.LEFT,
                    Direction.LEFT,
                    Direction.UP,
                    Direction.UP,
                    Direction.UP,
                ]
            )

        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.find_move_path(1, 42000)
            with pytest.raises(IndexError):
                board_graph.find_move_path(42000, 1)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.find_move_path(_, 1)
                with pytest.raises(exc_cls):
                    board_graph.find_move_path(1, _)

        def it_returns_empty_sequence_if_movement_is_blocked(self, board_graph):
            assert (
                board_graph.find_move_path(index_1d(11, 8, board_graph.board_width), 0)
                == []
            )

    class describe_positions_path_to_directions_path:
        def it_converts_path(self, board_graph, positions_path, directions_path):
            calculated_directions_path = board_graph.positions_path_to_directions_path(
                positions_path
            )
            assert calculated_directions_path == directions_path

        def it_raises_if_any_of_positions_are_invalid_values(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.positions_path_to_directions_path([42000])
            with pytest.raises(IndexError):
                board_graph.positions_path_to_directions_path([1, 42000])
            with pytest.raises(IndexError):
                board_graph.positions_path_to_directions_path([42000, 1])

            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.positions_path_to_directions_path([_])
                with pytest.raises(exc_cls):
                    board_graph.positions_path_to_directions_path([1, _])
                with pytest.raises(exc_cls):
                    board_graph.positions_path_to_directions_path([_, 1])

        def it_returns_empty_path_if_source_path_is_to_short(self, board_graph):
            assert board_graph.positions_path_to_directions_path([]) == []
            assert board_graph.positions_path_to_directions_path([1]) == []

    class describe_mark_play_area:
        def it_calculates_playable_area_of_board_marking_all_playable_cells(self):
            board_str = "\n".join(
                [
                    # 123456
                    "#######",  # 0
                    "#.$# @#",  # 1
                    "#######",  # 2
                    "#     #",  # 3
                    "#######",  # 4
                ]
            )
            board_graph = BoardGraph(SokobanPuzzle(board=board_str))

            expected_playable_cells = [
                index_1d(1, 1, 7),
                index_1d(2, 1, 7),
                index_1d(4, 1, 7),
                index_1d(5, 1, 7),
            ]

            board_graph.mark_play_area()
            for pos in range(board_graph.vertices_count):
                if pos in expected_playable_cells:
                    assert board_graph[pos].is_in_playable_area
                else:
                    assert not board_graph[pos].is_in_playable_area

    class describe_positions_reachable_by_pusher:
        board_str = "\n".join(
            [
                # 123456
                "#######",  # 0
                "#.$  @#",  # 1
                "# # ###",  # 2
                "#   #  ",  # 3
                "#####  ",  # 4
            ]
        )
        board_graph = BoardGraph(SokobanPuzzle(board=board_str))

        def it_returns_list_of_positions_reachable_by_pusher_movement_only(self):
            expected = [
                index_1d(5, 1, 7),
                index_1d(4, 1, 7),
                index_1d(3, 1, 7),
                index_1d(3, 2, 7),
                index_1d(3, 3, 7),
                index_1d(2, 3, 7),
                index_1d(1, 3, 7),
                index_1d(1, 2, 7),
                index_1d(1, 1, 7),
            ]
            assert (
                self.board_graph.positions_reachable_by_pusher(
                    pusher_position=index_1d(5, 1, 7)
                )
                == expected
            )

        def it_doesnt_require_that_start_position_actually_contain_pusher(self):
            expected = [
                index_1d(4, 1, 7),
                index_1d(3, 1, 7),
                index_1d(3, 2, 7),
                index_1d(3, 3, 7),
                index_1d(2, 3, 7),
                index_1d(1, 3, 7),
                index_1d(1, 2, 7),
                index_1d(1, 1, 7),
            ]
            assert (
                self.board_graph.positions_reachable_by_pusher(
                    pusher_position=index_1d(4, 1, 7)
                )
                == expected
            )

        def it_can_exclude_some_positions(self):
            expected = [
                index_1d(5, 1, 7),
                index_1d(4, 1, 7),
                index_1d(3, 1, 7),
                index_1d(3, 2, 7),
            ]
            excluded = [
                index_1d(3, 3, 7),
                index_1d(2, 3, 7),
                index_1d(1, 3, 7),
                index_1d(1, 2, 7),
                index_1d(1, 1, 7),
            ]
            assert (
                self.board_graph.positions_reachable_by_pusher(
                    pusher_position=index_1d(5, 1, 7), excluded_positions=excluded
                )
                == expected
            )

        def it_raises_if_start_position_is_illegal_value(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.positions_reachable_by_pusher(42000)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.positions_reachable_by_pusher(_)

    class describe_normalized_pusher_position:
        board_str = "\n".join(
            [
                # 123456
                "#######",  # 0
                "#.$  @#",  # 1
                "# # ###",  # 2
                "#   #  ",  # 3
                "#####  ",  # 4
            ]
        )
        board_graph = BoardGraph(SokobanPuzzle(board=board_str))

        def it_returns_top_left_position_of_pusher_in_his_reachable_area(self):
            assert self.board_graph.normalized_pusher_position(
                pusher_position=index_1d(5, 1, 7)
            ) == index_1d(1, 1, 7)

        def it_doesnt_require_that_start_position_actually_contain_pusher(self):
            assert self.board_graph.normalized_pusher_position(
                pusher_position=index_1d(4, 1, 7)
            ) == index_1d(1, 1, 7)

        def it_can_exclude_some_positions(self):
            assert self.board_graph.normalized_pusher_position(
                pusher_position=index_1d(4, 1, 7),
                excluded_positions=[index_1d(1, 1, 7)],
            ) == index_1d(3, 1, 7)

        def it_raises_if_start_position_is_illegal_value(self, is_using_native):
            with pytest.raises(IndexError):
                self.board_graph.normalized_pusher_position(42000)
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    self.board_graph.normalized_pusher_position(_)

    class describe_path_destination:
        def it_calculates_destination_position_from_source_and_directions_path(
            self, board_graph
        ):
            directions_path = [Direction.UP, Direction.RIGHT]
            start_position = index_1d(11, 8, board_graph.board_width)
            assert board_graph.path_destination(
                start_position, directions_path
            ) == index_1d(12, 7, board_graph.board_width)

        def it_silently_stops_search_on_first_off_board_position(self, board_graph):
            directions_path = [Direction.DOWN, Direction.DOWN, Direction.DOWN]
            start_position = index_1d(11, 8, board_graph.board_width)
            assert board_graph.path_destination(
                start_position, directions_path
            ) == index_1d(11, 10, board_graph.board_width)

        def it_silently_stops_search_on_illegal_direction(self, board_graph):
            directions_path = [Direction.DOWN, Direction.NORTH_WEST]
            start_position = index_1d(11, 8, board_graph.board_width)
            assert board_graph.path_destination(
                start_position, directions_path
            ) == index_1d(11, 9, board_graph.board_width)

        def it_raises_if_start_position_is_illegal_value(
            self, is_using_native, board_graph
        ):
            with pytest.raises(IndexError):
                board_graph.path_destination(42000, [])
            exc_cls = TypeError if is_using_native else KeyError
            for _ in [-42000, None, "", "ZOMG!", "42"]:
                with pytest.raises(exc_cls):
                    board_graph.path_destination(_, [])

    # class describe__reachables:
    #     board = "\n".join(
    #         [
    #             # 123456
    #             "#######",  # 0
    #             "#.$# @#",  # 1
    #             "#######",  # 2
    #             "#     #",  # 3
    #             "#######",  # 4
    #         ]
    #     )
    #     board_graph = BoardGraph(SokobanPuzzle(board=board))

    #     def it_calculates_all_positions_reachable_from_root(self):
    #         if not hasattr(self.board_graph, "_reachables"):
    #             return

    #         root = index_1d(5, 1, 7)
    #         assert self.board_graph._reachables(root) == [root, index_1d(4, 1, 7)]

    #     def it_skips_explicitly_excluded_positions(self):
    #         if not hasattr(self.board_graph, "_reachables"):
    #             return

    #         root = index_1d(5, 1, 7)
    #         assert self.board_graph._reachables(root, excluded_positions=[root]) == [
    #             index_1d(4, 1, 7)
    #         ]
    #         root = index_1d(5, 1, 7)
    #         assert self.board_graph._reachables(
    #             root, excluded_positions=[index_1d(4, 1, 7)]
    #         ) == [root]

    class describe_reconfigure_edges:
        def it_reconfigures_all_edges_in_board(self):
            graph = BoardGraph(SokobanPuzzle(width=2, height=2))
            graph.remove_all_edges()
            graph.reconfigure_edges()

            assert graph.edges_count == 8
            assert graph.has_edge(0, 1, Direction.RIGHT)
            assert graph.has_edge(1, 0, Direction.LEFT)
            assert graph.has_edge(0, 2, Direction.DOWN)
            assert graph.has_edge(2, 0, Direction.UP)
            assert graph.has_edge(2, 3, Direction.RIGHT)
            assert graph.has_edge(3, 2, Direction.LEFT)
            assert graph.has_edge(1, 3, Direction.DOWN)
            assert graph.has_edge(3, 1, Direction.UP)

        def it_doesnt_create_duplicate_direction_edges_in_multidigraph(self):
            graph = BoardGraph(TriobanPuzzle(width=2, height=2))
            graph.remove_all_edges()
            graph.reconfigure_edges()
            assert graph.out_edges_count(0, 1) == 2
            assert graph.out_edges_count(1, 0) == 2
